
unit  TedIdleControl;

interface
uses  Windows, SysUtils, Classes, Forms;

type  TTedIdleOutlet = class (TObject)
      private
        FOnIdle:    TIdleEvent;
        FPriority:  Integer;
        FPassCount: Integer;
        FIsBusy:    Boolean;
        FEnabled:   Boolean;
        FDelay:     Integer;
        procedure InitSchedulerEntry;
        procedure SetOnIdle   (const Handler:  TIdleEvent);
        procedure SetPriority (const Priority: Integer);
        procedure SetDelay    (const Delay:    Integer);
      public
        property OnIdle: TIdleEvent read FOnIdle   write SetOnIdle;
        property Enabled:   Boolean read FEnabled  write FEnabled;
        property Priority:  Integer read FPriority write SetPriority;
        property Delay:     Integer read FDelay    write SetDelay;
        constructor Create;
        destructor Destroy; override;
      end;

      TTedIdleControl = class (TComponent)
      private
        FOutlet: TTedIdleOutlet;
        function  GetOnIdle:      TIdleEvent;
        function  GetEnabled:     Boolean;
        function  GetPriority:    Integer;
        function  GetDelay:       Integer;
        function  GetGlobalDelay: Integer;
        procedure SetOnIdle      (const Handler:  TIdleEvent);
        procedure SetEnabled     (const Enabled:  Boolean);
        procedure SetPriority    (const Priority: Integer);
        procedure SetDelay       (const Delay:    Integer);
        procedure SetGlobalDelay (const Delay:    Integer);
      published
        property OnIdle:  TIdleEvent read GetOnIdle      write SetOnIdle;
        property Enabled:    Boolean read GetEnabled     write SetEnabled     default TRUE;
        property Priority:   Integer read GetPriority    write SetPriority    default 1;
        property Delay:      Integer read GetDelay       write SetDelay       default 0;
        property SlowMotion: Integer read GetGlobalDelay write SetGlobalDelay default 0;
      public
        constructor Create (AOwner : TComponent); override;
        destructor Destroy;                       override;
      end;


procedure Register;
procedure SetSlowMotion    (const Delay:     Integer);
procedure SetIdleInstall   (const Automatic: Boolean);
procedure TriggerIdleEvent (const Sender: TObject; var Done: Boolean);


{****** Documentation *********************************************************

This is the "Heinzelmnnchen"-Control :-)
(Heinzelmnnchen are people from the local legends of Cologne looking like
 small dwarfs which do the work while people sleep.)


Classes:    TTedIdleControl, TTedIdleOutlet
Properties: OnIdle, Enabled, Priority, Delay, SlowMotion

An idle handler can be used to perform tasks while waiting for input. Delphi
supports one global idle handler (Application.OnIdle), while we offer a local
idle handler, by TTedIdleControl for visual programming, and by TTedIdleOutlet
to give you access to the bare functionality.

Advantages of our local idle handling:
+ as many idle handlers as you need
+ plug-in your handlers interactively using the form designer
+ or integrate outlets for idle handlers into your objects
+ subsumes global idle handling
+ easy upgrade from global idle handling
+ easy scheduling of multiple handlers by priorities
+ full control over the interfacing to Delphi
+ delays and "slow motion" for easy testing

HINT: If you are already familiar to Delphi's idle handling and just need to
      have your handlers called once when the application becomes idle, just
      read "How to use local idle handling". If you do not touch the parameter
      Done, then anything will work fine without deeper understanding and you
      do not need to worry about priorities.

How Delphi supports global idle handling:

Idle handling refers to the message loop of an application. An interactive
program runs by repeatedly checking the message queue of the application and
processing the messages. These messages report keyboard and mouse input and
any other system requests or notifications. When there is no message left in
the queue, the application is idle and processes the OnIdle event. If there
is still no message in the queue after executing the event handler, the
application calls WaitMessage to yield control to other applications until
another message arrives. This improves the all-over performance in
multi-tasking environments. The idle handler can cause the application not to
call WaitMessage but instead to call the idle handler again.

The boolean Done parameter of an idle handler:

It is TRUE by default. Keeping this signals that there is nothing else to do
for this handler while the application is idle, and this handler does not
object to calling WaitMessage. If the handler sets it to FALSE, it signals
that this handler should be called again when there is no message to be
processed. This prevents the application from calling WaitMessage before
calling this handler again.

You can use Done to perform tasks in small steps while the application is
idle. Each time the idle handler is called, you execute one step. This way
a long task can be processed by an idle handler without locking up the main
thread, so message handling always goes first. To use this slicing technique
you need to keep track of the state your task is in, so you can set Done to
FALSE while the task proceeds, and leave Done at TRUE when the task is
finished.

How to use local idle handling:

You must decide whether to use global or local idle handling (you can NOT use
both). To use Delphi's global idle handling, install a handler in
Application.OnIdle. To use our local idle handling, just install a handler in
an idle control, of which you can create as many as you require. Additionally,
you can install handlers not using controls, via TTedIdleOutlet objects. All
these local handlers will be served then nearly as if they were installed
alone. If you worry about "nearly", consult "How local idle handling works".
You can balance problems by using priorities.

NOTE: You can not use Application.OnIdle, if you install any local idle
      handler, because in that case this unit occupies it. Install the required
      global handler by using a TTedIdleControl or a TTedIdleOutlet.

      If you need our unit not to install into Application.OnIdle
      automatically, it can be switched off. See "How to control use of
      Application.OnIdle".

Disabling an idle handler

There is a property Enabled to be used at run time to temporarily switch off
the calling of an installed handler. You can initialize it at design time.

The difference to deinstalling the handler (OnIdle := nil) is a matter of
timing. Switching Enabled is fast, for the idle control or outlet remains in
the managing structures, and scheduling (see "How local idle handling works")
continues. Just the call to your handler is pruned. So using Enabled is
recommended if you wish to turn the handler off for a few cycles or if you
need the scheduling to proceed undisturbed. Deinstalling and reinstalling
requires some time to change the tables and scheduling is reset for this
handler. But it has the advantage of no longer consuming time (for scheduling)
on Delphi's idle event. Because the idle control or outlet becomes invisible
then to the idle scheduler effectively, this is the method of choice to remove
an idle handler for a longer period.

How to install local idle handlers not attached to controls

You can create and destroy TTedIdleOutlet objects at any time. They contain
just the bare functionality and carry no stuff for visual programming. You
can install your idle handler there as at an idle control. This can be useful
to create idle outlets at run time as well as to build an handler into another
object. Carrying an TTedIdleOutlet has no effect except consuming the memory
occupied by it while no handler is installed there.

Upgrading from global idle handling:

If you wish to use local idle handling but have a global handler in use, just
install the old handler at one idle control. If you do not install any other
idle handler, anything behaves as before. Then you can add other local
handlers, which are served as well. If your old handler relies on the
implementation of Delphi's Application.OnIdle, or if you like to use Done in
odd ways, you might need to balance things by using priorities, but there will
be almost no case in which you have to change your old code.

How to control use of Application.OnIdle

Normally, our unit occupies Application.OnIdle at the very moment the first
handler is installed in an idle control or outlet, and releases it when it is
no longer needed. SetIdleInstall (FALSE) switches this behavior off and
SetIdleInstall (TRUE) turns it on again. Application.OnIdle is released or
occupied as appropriate, so SetIdleInstall can be used at any time.

If you inhibit automatic occupation of Application.OnIdle, you have to trigger
our unit explicitly. To compensate for the no longer incoming Delphi event, you
have to call TriggerIdleEvent as appropriate. Note: You can use priority 0 to
get a local handler called on every Delphi OnIdle event, so you do not need to
sacrifice the automatic install for that.

How local idle handling works:

Our local idle handling is build on top of Delphi's global idle handling.
You should be familiar to Delphi's idle handling to understand this section
(refer to "How Delphi supports global idle handling" if necessary). Delphi's
OnIdle event is used to trigger the local OnIdle events.

Imagine the set of local idle handlers, and none of them setting Done to FALSE.
Then, when Delphi's OnIdle event occurrs, one local handler is selected and
executed. All handlers are served successively triggered by successive Delphi
events, before we allow the application to call WaitMessage. This ensures that
(1) during Delphi's idle event no more time is consumed as if there were only
one handler installed, so message handling is not delayed, and
(2) all handlers are served before the application yields control to another.

To proceed to the other cases where Done is set to FALSE, you must have a
thorough understanding of that parameter (refer to "The boolean Done parameter
of an idle handler" in case of doubt).

If any local idle handler sets Done to FALSE while we are performing one
round over all handlers, we do not allow the application to call WaitMessage
after we have finished the round, but we begin another round. This ensures
that a handler will be called again prior to WaitMessage if that handler
requires it. So handlers doing some slicing are served appropriately: till they
are ready. All other handlers (which returned with Done having the value TRUE)
are called again as well. This should not be a problem, because a handler
normally checks for work to do rather than doing big work just for fun, or
otherwise it would slow down the application, even if it were alone. One might
be tempted to demand excluding the already done handlers from the next round.
But this would lead to starvation, if any handler never returns with Done being
TRUE, or would be unfair if one handler takes long to arrive there. Using
priorities you can trade the frequency of service for pressing down the
overhead which is caused by this safe strategy.

How to balance problems by using priorities

You need only to worry about priorities if any local idle handler sets Done
to FALSE, or if you need some handlers being called on every Delphi OnIdle
event. (Read the previous section to understand the following.)

In a situation where we can not allow the application to call WaitMessage,
because some local idle handler signaled need of continuosly being called,
all handlers must be served eventually to avoid starvation. But the programmer
has knowledge about the required frequency of service. Imagine one local
handler which updates a display in steps as soon as processor capacity is
available, because the whole task takes long and thus can not be performed in
the paint method of this display. And imagine the other handlers do some small
monolithic things. It would slow down the display update to call all the funny
small handlers on every update step, but nevertheless they must be served on
on a regular basis. Let's call them, say, on every 10th step only. If you wish
a certain handler to be called every 10th time only while there is a busy
handler, set the priority of that certain handler to 10.

When the application becomes idle the first time or after a WaitMessage, all
local idle handlers are processed in the first round, so every handler gets
its chance. If anyone of them requires another round, scheduling begins. In
round N a handler with priority P is called, iff P MOD N = 1 (the first round
has number 1). After a round in which no handler requires another round and
there is no busy handler scheduled for a later round, scheduling is aborted
and WaitMessage is called. Note that messages might be processed while
scheduling goes on. This strategy ensures that
(1) all handlers are satisfied before the application yields control, and
(2) no scheduling delay occurrs when there is no busy handler.

It is allowed to change the priority of a handler at any time, even from
within this handler, and it is considered to be fast. During scheduling
changing the priority of a handler sets the round in which the handler runs
next with respect to the rounds already waited. This technique allows the
implementation of more complex scheduling strategies if required.

Special case: If you need a handler being called each time Delphi's OnIdle
event occurrs, set your handlers priority to 0 to achive that. On the event,
all priority 0 handlers are executed, then one of the other handlers is
selected to run. Use this only if you have no other chance, because calling
more than one handler on the event is an overkill in most cases and might be
a problem.

Testing with delays and slow motion

To ensure the correct behavior of your application it is often useful to
insert delays into some idle handler and watch the things going on. Imagine
an idle handler which paints a text window one line per event while retrieving
data from a database. On small test data it might run much to fast to watch,
but it can be essential that it works when dealing with huge data. To check
the behavior set the Delay property of the respective idle handler to a value
greater than zero, for example 500. Then, the execution of the program will
be delayed after the handler is executed, for the given number of milliseconds.
Now you can watch the lines being written one by one.

You can delay each idle handler seperately by its delay property to check it
independently. This also allows to try out the effect of setting some
properties, if you set the delays, enlarged by a big factor, to the estimated
time consumed by the respective handler. Another mode of testing is to use
slow motion. By the property SlowMotion, you can set a global delay common
to all idle handlers, which is inserted after Delphi's OnIdle event is
processed (= after any scheduled idle handler is processed). This enables
you to watch the interplay of the handlers, especially when using priorities.

The delay is done by the Windows operation Sleep, which just blocks the
calling thread as if it were busy, but without consuming processing power.
This has the advantage of not interfering with the message handling of your
application, but the processing of input is delayed, too. Your application
will respond slowly while being tested.

Don't forget to set SlowMotion and all Delay values back to zero, before
you generate the final version of your program.

Implementation note

Delay and SlowMotion are implemented using Windows.Sleep. This should be
no problem, for Sleep is available under Windows95, WindowsNT and Win32s.
No other Windows operation is used directly.

******************************************************************************}


implementation

const MaxPriority = 9999;

type  TEntertainer = class
        FOutlets:      TList;
        FEntryCount,
        FZeroCount,
        FBusyCount:    Integer;
        FIsFirstRound: Boolean;
        FCurrent,
        FExecuting:    Integer;
        FLiving:       Boolean;
        constructor Create;
        destructor Destroy; override;
        procedure Install;
        procedure Deinstall;
        procedure IncrementZeroCount (const Delta: Integer);
        procedure Insert (const Outlet: TTedIdleOutlet);
        procedure Delete (const Outlet: TTedIdleOutlet);
        procedure InitScheduler;
        procedure Execute   (const Sender: TObject);
        procedure ZeroIdleHandler (Sender: TObject; var Done: Boolean);
        procedure     IdleHandler (Sender: TObject; var Done: Boolean);
        function Scheduled  (const ThisRoundOnly: Boolean):   Boolean;
      end;

var   Entertainer:      TEntertainer = nil;
      Trigger:          TIdleEvent   = nil;
      AutomaticInstall: Boolean      = TRUE;
      GlobalDelay:      Integer      = 0;       


procedure Register;
begin
  RegisterComponents ('TEDSoft', [TTedIdleControl]);
end;


procedure SetSlowMotion (const Delay: Integer);
begin
  if         Delay >= 0
  then GlobalDelay := Delay
end;


procedure SetIdleInstall (const Automatic: Boolean);
begin
  if   Automatic <> AutomaticInstall
  then begin
         AutomaticInstall := Automatic;
         if   Automatic
         then Application.OnIdle := Trigger
         else Application.OnIdle := nil
       end
end;


procedure TriggerIdleEvent (const Sender: TObject; var Done: Boolean);
begin
  if   Assigned (Trigger)
  then Trigger  (Sender, Done)
end;


constructor TEntertainer.Create;
begin
  inherited Create;
  FOutlets    := TList.Create;
  FEntryCount := 0
end;


destructor TEntertainer.Destroy;
begin
  if   FEntryCount <> 0
  then Deinstall;
  FOutlets.Free;
  inherited Destroy
end;


procedure TEntertainer.Install;
begin
  if   FZeroCount <> 0
  then Trigger := ZeroIdleHandler
  else Trigger :=     IdleHandler;
  if   AutomaticInstall
  then Application.OnIdle := Trigger
end;


procedure TEntertainer.Deinstall;
begin
  Trigger := nil;
  if   AutomaticInstall
  then Application.OnIdle := nil
end;


procedure TEntertainer.IncrementZeroCount (const Delta: Integer);
var OldZeroCount: Integer;
begin
  if   Delta <> 0
  then begin
         OldZeroCount := FZeroCount;
         Inc (FZeroCount, Delta);
         if   (FEntryCount <> 0) AND ((OldZeroCount = 0) <> (FZeroCount = 0))
         then Install
       end
end;


procedure TEntertainer.Insert (const Outlet: TTedIdleOutlet);
begin
  if FEntryCount      = 0 then InitScheduler;
  if Outlet.FPriority = 0 then IncrementZeroCount (1);
  if FEntryCount      = 0 then Install;                                     {!}
  FOutlets.Add (Outlet); Inc (FEntryCount)
end;


procedure TEntertainer.Delete (const Outlet: TTedIdleOutlet);
var j: Integer;
begin
  j := FOutlets.Remove (Outlet); Dec (FEntryCount);
  if FEntryCount      = 0 then Deinstall;                                   {!}
  if Outlet.FPriority = 0 then IncrementZeroCount (-1);
  if Outlet.FIsBusy       then Dec (FBusyCount);
  if j  = FExecuting      then FLiving := FALSE;
  if j <= FExecuting      then Dec (FExecuting);
  if j <= FCurrent        then Dec (FCurrent)
end;


procedure TEntertainer.InitScheduler;
begin
  FBusyCount    := 0;
  FZeroCount    := 0;
  FIsFirstRound := TRUE;
  FCurrent      := 0;
  FExecuting    := -1
end;


procedure TEntertainer.Execute (const Sender: TObject);
{number of entry to be executed must be in FExecuting}
var Ok, OldIsBusy: Boolean;
begin
  with TTedIdleOutlet (FOutlets [FExecuting]) do
  begin
    FPassCount := 0;
    if   FEnabled
    then begin
           FLiving   := TRUE;
           OldIsBusy := FIsBusy;
           Ok        := TRUE;
           FOnIdle (Sender, Ok);
           if   FLiving
           then begin
                  FIsBusy := NOT Ok;
                  Inc (FBusyCount, Ord (FIsBusy) - Ord (OldIsBusy))
                end;
           if   FDelay > 0
           then Sleep (FDelay)
         end
  end
end;


procedure TEntertainer.ZeroIdleHandler (Sender: TObject; var Done: Boolean);
begin
  {execute all priority 0 handlers}
  FExecuting := 0;
  while FExecuting < FEntryCount do
  with TTedIdleOutlet (FOutlets [FExecuting]) do
  begin
    if   FPriority = 0
    then Execute (Sender);
    Inc (FExecuting)
  end;

  {execute one normal handler}
  IdleHandler (Sender, Done)
end;


procedure TEntertainer.IdleHandler (Sender: TObject; var Done: Boolean);
begin
  {execute one normal handler}
  if   Scheduled (FALSE)
  then begin
         FExecuting := FCurrent;
         Execute (Sender);
         Inc (FCurrent);
         if   Scheduled (TRUE) OR (FBusyCount <> 0)
         then Done := FALSE
       end
  else begin
         if   FBusyCount <> 0
         then Done := FALSE
       end;

  {perform slow motion}
  if   GlobalDelay > 0
  then Sleep (GlobalDelay)
end;


function TEntertainer.Scheduled (const ThisRoundOnly: Boolean): Boolean;
{number of entry to be scheduled must be in FCurrent and is modified}
begin
  Result := FALSE;
  if   FEntryCount > FZeroCount {there is a normal handler}
  then begin
         repeat
           if   FCurrent >= FEntryCount {a round is finished}
           then begin
                  if   ThisRoundOnly
                  then break;
                  {start a new round}
                  FIsFirstRound := FBusyCount = 0;
                  FCurrent      := 0
                end;
           with TTedIdleOutlet (FOutlets [FCurrent]) do
           begin
             if   FPriority <> 0
             then begin
                    if   FIsFirstRound
                    then begin                   Result := TRUE                    end
                    else begin Inc (FPassCount); Result := FPassCount >= FPriority end
                  end
           end;
           if   Result
           then break;
           Inc (FCurrent)
         until FALSE
       end
end;


constructor TTedIdleOutlet.Create;
begin
  inherited Create;
  FOnIdle   := nil;
  FEnabled  := TRUE;
  FPriority := 1
end;


destructor TTedIdleOutlet.Destroy;
begin
  OnIdle := nil;
  inherited Destroy
end;


procedure TTedIdleOutlet.InitSchedulerEntry;
begin
  FPassCount := MaxPriority; {for the first time run as soon as possible}
  FIsBusy    := FALSE
end;


procedure TTedIdleOutlet.SetOnIdle (const Handler: TIdleEvent);
var  IsActive,
    WasActive: Boolean;
begin
   IsActive := Assigned (Handler);
  WasActive := Assigned (FOnIdle);
  FOnIdle   := Handler;
  if   IsActive
  then InitSchedulerEntry;
  if   IsActive <> WasActive
  then begin
         if   IsActive
         then begin
                if   Entertainer = nil
                then Entertainer := TEntertainer.Create;
                Entertainer.Insert (Self)
              end
         else begin
                if   Entertainer <> nil
                then Entertainer.Delete (Self)
              end
       end
end;


procedure TTedIdleOutlet.SetPriority (const Priority: Integer);
begin
  if   (Priority >= 0) AND (Priority <= MaxPriority)
  then begin
         if   Assigned (FOnIdle) AND (Entertainer <> nil)
         then Entertainer.
                IncrementZeroCount (Ord (Priority = 0) - Ord (FPriority = 0));
         FPriority := Priority
       end
end;


procedure TTedIdleOutlet.SetDelay (const Delay: Integer);
begin
  if    Delay >= 0
  then FDelay := Delay
end;


constructor TTedIdleControl.Create (AOwner : TComponent);
begin
  inherited Create (AOwner);
  FOutlet := TTedIdleOutlet.Create
end;


destructor TTedIdleControl.Destroy;
begin
  FOutlet.Free;
  inherited Destroy
end;


function TTedIdleControl.GetOnIdle: TIdleEvent;
begin
  Result := FOutlet.OnIdle
end;


procedure TTedIdleControl.SetOnIdle (const Handler: TIdleEvent);
begin
  FOutlet.OnIdle := Handler
end;


function TTedIdleControl.GetEnabled: Boolean;
begin
  Result := FOutlet.Enabled
end;


procedure TTedIdleControl.SetEnabled (const Enabled:  Boolean);
begin
  FOutlet.Enabled := Enabled
end;


function TTedIdleControl.GetPriority: Integer;
begin
  Result := FOutlet.Priority
end;


procedure TTedIdleControl.SetPriority (const Priority: Integer);
begin
  FOutlet.Priority := Priority
end;


function TTedIdleControl.GetDelay: Integer;
begin
  Result := FOutlet.Delay
end;


procedure TTedIdleControl.SetDelay (const Delay: Integer);
begin
  FOutlet.Delay := Delay
end;


function TTedIdleControl.GetGlobalDelay: Integer;
begin
  Result := GlobalDelay
end;


procedure TTedIdleControl.SetGlobalDelay (const Delay: Integer);
begin
  SetSlowMotion (Delay)
end;


initialization
finalization
  Entertainer.Free
end.
